// Name: main.c
// Project: Smart-Response-PE/examples/hello-world-clicker by Author: Stephen Erisman <github@serisman.com> Creation Date: 2018-09-21
// Hacked relentlessly by M. Ray Burnette 20230613

#include "Arduino.h"					// Implement the construct of setup() and loop()
// #include <stdlib.h>					// abs, labs, atof, atoi, atol, _uitoa, _itoa, _ultoa, _ltoa. RAND_MAX, rand, srand
#include "hal.h"						// #include "cc253x/cc253x.h"
#include "util.h"						// Compile macros
#include "string_utils.h"				// prototypes for all *_str.c functions
#include "clock.h"						// Setup for millis and micros via Interrupt
#include "display.h"					// buffer screen[NUM_SCREEN_ROWS * SCREEN_WIDTH]
#include "keypad.h"						// keymap array to button-by-name defines, keypad_scan, keypad_poll, etc

void timeSincePwrOn(void);				// function prototypes / forward declarations
void nextFrame(void);
void moveDigitsRight(void);
void moveDigitsLeft(void);
void displayZero(void);
void displayShow(void);
void PolishStackPush(void);
void DoTheMath(void);
void displayMathOp(void);
void StackRollUp(void);
void StackRollDn(void);
void showRegAsDecimal(float reg, int x, int y);

#define FPS	30u							// LCD screen buffer update
/*
SDCC data types: char (8 bits, 1 byte), short (16 bits, 2 bytes), 
				int (16 bits, 2 bytes), long (32 bit, 4 bytes), 
				long long (64 bit, 8 bytes), float (4 byte IEEE),
				and _Bool/bool.
*/

// __data indication to SDCC to use "fast SRAM" where __xdata is generic SRAM
char		__xdata str_buf[12];
uint32_t 	__xdata millis;
int8_t 		__xdata scroll_x = 0;
int8_t 		__xdata scroll_x_dir = 1;
int8_t 		__xdata scroll_y = 0;
int8_t 		__xdata scroll_y_dir = 1;
int8_t		__xdata digits[10];
int8_t		__xdata decade = 0;
float    	__xdata X_reg = 0;			// X, Y, Z, T like on a reverse-polish HP
float    	__xdata Y_reg = 0;
float    	__xdata Z_reg = 0;
float    	__xdata T_reg = 0;
float    	__xdata Tmp_reg = 0;		// Last X,  temporary
float		__xdata Tmp_float = 0;		// for X display as decimal: (float - int) * 1000 = 3 decimal places
bool		__xdata l_more = false;		// logical-more == characters available
unsigned long t_now;					// event timer
int8_t		DoDefault = 0;


void setup() {
	oscillator_32mhz();					// defined in oscillator.h
	clock_init();
	ENABLE_INTERRUPTS
	display_init();
	display_set_frame_rate(FPS);
	displayZero();						// initialize keyboard buffer for digits
	/*
	display_set_cursor(40,10);			// (# Pixels down, # Pixels right) all referenced to (0,0) top-left corner
	display_print("Ray's RPN");
	display_set_cursor(40, 20);
	display_print("v 20130613");
	*/
}

void loop() {
	if (display_next_frame()) {			// Wait until its time to render our next frame
		if (decade < 8) {				// max digits in array? "99999999" gives FP math overflow, limit to 7 digits input
			l_more = true;
		} else {
			l_more = false;
		}
		t_now = clock_millis();
		nextFrame();
	}
}

void nextFrame() {
	keypad_poll();						// Poll the keypad
	if (keypad_any_pressed())
	{
		for (uint8_t col=0; col<4; col++) {
		  for (uint8_t row=0; row<8; row++) {
			uint8_t mask = util_bit_to_mask[row];
			if (keypad_pressed(col, mask)) {
			  uint32_t Results = (uint32_t)(row * 5 + col);
			  
			  switch (Results)
				{
				case 0:					// Enter == push stack
					PolishStackPush();
					displayZero(); 
					if (Tmp_reg > 0) Tmp_reg = 0;
					break;
				case 1:					// Select Arithemetic Operation: + - / * Lf arrow 					
				if (DoDefault == 0) {
						DoDefault = 3;	// roll-over
					} else {
						DoDefault--;
					}
					break;
				case 2:					// Rt arrow "x" Excute arithemetic
					if(Tmp_reg != 0) PolishStackPush(); 
					DoTheMath();
					Tmp_reg = 0;
					displayZero();
					break;
				case 3:					// del/Backspace
					if (decade > 0) {
						moveDigitsLeft();
						decade = decade--;
						digits[decade] = 0;
					}
					break;
				case 5:					// ?/Help	display time since PowerOn
					timeSincePwrOn();
					break;
				case 6:					// A/1
					if (l_more) {moveDigitsRight(); digits[0] = 1; ++decade;}
					break;
				case 7:					// B/2
					if (l_more) {moveDigitsRight(); digits[0] = 2; ++decade;}
					break;
				case 8:					// C/3
					if (l_more) {moveDigitsRight(); digits[0] = 3; ++decade;}
					break;
				case 10:				// Menu/Home == Clear the digit buffer and Tmp value
					displayZero();
					break;
				case 11:				// D/4
					if (l_more) {moveDigitsRight(); digits[0] = 4; ++decade;}
					break;
				case 12:				// E/5
					if (l_more) {moveDigitsRight(); digits[0] = 5; ++decade;}
					break;
				case 13:				// F/6
					if (l_more) {moveDigitsRight(); digits[0] = 6; ++decade;}
					break;
				case 15:				// Up
					StackRollUp();
					break;
				case 16:				// G/7
					if (l_more) moveDigitsRight(); digits[0] = 7; ++decade;
					break;
				case 17:				// H/8
					if (l_more) moveDigitsRight(); digits[0] = 8; ++decade;;
					break;
				case 18:				// I/9
					if (l_more) {moveDigitsRight(); digits[0] = 9; ++decade;}
					break;
				case 20:				// Down
					StackRollDn();
					break;
				case 21:				// +/-   negate X_reg
					X_reg = X_reg * -1.0;
					break;
				case 22:				// J/0
					if (l_more) {moveDigitsRight(); digits[0] = 0; ++decade;}
					break;
				case 23:				// x-y/.	rotate x-y
					Tmp_reg = X_reg; X_reg = Y_reg; Y_reg = Tmp_reg; Tmp_reg = 0;
					break;
				case 35:				// Power
					break;
				default:
					break;
				}
			}
		  }
		}
		clock_delay_ms(180);			// debounce: elastomer keyboard kind of sucks
	}
	displayShow();	
}

void moveDigitsRight(void) {			// Move Right digits [0] ALWAYS first # typed
	digits[9] = digits[8] ; digits[8] = digits[7] ; digits[7] = digits[6] ;
	digits[6] = digits[5] ;	digits[5] = digits[4] ; digits[4] = digits[3] ;
	digits[3] = digits[2] ; digits[2] = digits[1] ; digits[1] = digits[0] ; 
}

void moveDigitsLeft(void) {				// move displayed digits to the Left
	digits[0] = digits[1] ; digits[1] = digits[2] ; digits[2] = digits[3] ;
	digits[3] = digits[4] ;	digits[4] = digits[5] ;	digits[5] = digits[6] ;
	digits[6] = digits[7] ; digits[7] = digits[8] ; digits[8] = digits[9] ;
}

void displayZero(void) {
	for (int8_t j =0; j < 10; j++) {
		digits[j] = 0 ;
	}
	decade = 0; l_more = true;
	displayShow();
}

void displayShow(void) {
	//int x_cord, y_cord;
	Tmp_reg = digits[0]  \
		+ digits[1] * 10L \
		+ digits[2] * 100L \
		+ digits[3] * 1000L \
		+ digits[4] * 10000L \
		+ digits[5] * 100000L \
		+ digits[6] * 1000000L \
		+ digits[7] * 10000000L \
		+ digits[8] * 100000000L \
		+ digits[9] * 1000000000L;

		display_set_cursor(20,0);	display_print("T=");	showRegAsDecimal(T_reg, 40, 0);
		display_set_cursor(20,10);	display_print("Z=");	showRegAsDecimal(Z_reg, 40, 10);
		display_set_cursor(20,20);	display_print("Y=");	showRegAsDecimal(Y_reg, 40, 20);
		display_set_cursor(20,30);	display_print("X=");	showRegAsDecimal(X_reg, 40, 30);
		display_set_cursor(10,30);	displayMathOp();	// the mathematical operation for Y, X
		u32_to_str(str_buf, Tmp_reg);
		display_set_cursor(46,40);		// input buffer display position is last line shown
		if (Tmp_reg != 0) display_print(str_buf);
		// in Display.h macro: inline void display_paint(bool clear)   { uc1701_paint(screen, clear); }	
		display_paint(true);
}

void showRegAsDecimal(float reg, int x_cord, int y_cord ) {
	bool l_Flag = false;
	if (reg < 0) {
		reg = reg * -1.0;
		x_cord = x_cord - 5;
		l_Flag = true;
	}
	display_set_cursor(x_cord, y_cord);
	long Temp = (long) reg;				// integer component
	Tmp_float = (reg - Temp) *100000L;	// decimal component 5 digits
	u32_to_str(str_buf, reg);
	if (l_Flag) {
		display_print(" -"); 
	}	else {
		display_print(" ");
	}
	display_print(str_buf);
	display_print(".");
	u32_to_str(str_buf, Tmp_float);
	display_print(str_buf);
}

void PolishStackPush(void) {			// x --> y --> z --> t --> lost
	T_reg = Z_reg; Z_reg = Y_reg; Y_reg = X_reg; X_reg = Tmp_reg;
	Tmp_reg = 0;
}

void StackRollDn(void) {				// t --> z --> y --> x --> t & roll
	Tmp_reg = X_reg; X_reg = Y_reg; Y_reg = Z_reg; Z_reg = T_reg; T_reg = Tmp_reg;
}

void StackRollUp(void) {				// t --> x --> y --> z --> t & roll
	Tmp_reg = T_reg; T_reg = Z_reg; Z_reg = Y_reg; Y_reg = X_reg; X_reg = Tmp_reg;
}

void DoTheMath(void) {
	switch(DoDefault) {
		case 0:							// +	Y+X
			X_reg = (Y_reg + X_reg);
			break;
		case 1:							// -	Y-X
			X_reg = (Y_reg - X_reg);
			break;
		case 2:							// *	Y*X
			X_reg = (Y_reg * X_reg);
			break;
		case 3:							// /	Y/X
			X_reg = (Y_reg / X_reg);
			break;
		default:
			break;
	}
	// HP paradigm is to copy "T" down to lower level dropping stack
	Y_reg = Z_reg; Z_reg = T_reg; Tmp_reg = 0;
	displayShow();
}

void displayMathOp(void) {
	switch(DoDefault) {
		case 0:							// +	Y+X
			display_print("+");
			break;
		case 1:							// -	Y-X
			display_print("-");
			break;
		case 2:							// *	X*Y
			display_print("*");
			break;
		case 3:							// /	Y/X
			display_print("/");
			break;
		default:
			break;
	}
}

void timeSincePwrOn() {
	display_set_cursor(30 , 40);
	unsigned long currentMillis = clock_millis();
	display_set_cursor(30, 40);
	display_print("                    ");
	display_set_cursor(30, 40); display_print("UpTime ");
	unsigned long seconds = currentMillis / 1000;
	unsigned long minutes = seconds / 60;
	unsigned long hours = minutes / 60;
	unsigned long days = hours / 24;
	currentMillis %= 1000;
	seconds %= 60;
	minutes %= 60;
	if (hours < 10) display_print("0");
	u32_to_str(str_buf, hours); display_print(str_buf); display_print(":");
	if (minutes < 10) display_print("0"); 
	u32_to_str(str_buf, minutes); display_print(str_buf); display_print(":");
	if (seconds < 10) display_print("0");
	u32_to_str(str_buf, seconds); display_print(str_buf); display_print("");
}
